/*
 *  stub.c version 0.3, 2020-01-30
 *
 *  Copyright (c) 2019-2020 Public Software Group e. V., Berlin, Germany
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

/*
 *  This program calculates the lengths of stub transmission lines for radio
 *  applications where a low standing wave ratio over a broader (and/or split)
 *  range of frequencies is desired.
 *
 *  The program can be compiled by entering the following command:
 *    cc -Wall -o stub stub.c -lm
 */


/* Version string and copyright notice for binaries */

const char *version = "stub version 0.3, 2020-01-30";
const char *copyright = "Copyright (c) 2019-2020 Public Software Group e. V., Berlin, Germany";
const char *license = "\
Permission is hereby granted, free of charge, to any person obtaining a\n\
copy of this software and associated documentation files (the \"Software\"),\n\
to deal in the Software without restriction, including without limitation\n\
the rights to use, copy, modify, merge, publish, distribute, sublicense,\n\
and/or sell copies of the Software, and to permit persons to whom the\n\
Software is furnished to do so, subject to the following conditions:\n\
\n\
The above copyright notice and this permission notice shall be included in\n\
all copies or substantial portions of the Software.\n\
\n\
THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n\
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n\
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n\
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n\
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING\n\
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER\n\
DEALINGS IN THE SOFTWARE.\n";


/* Help text */

const char *help = "\
This program calculates the lengths of stub transmission lines for radio\n\
applications where a low standing wave ratio over a broader (and/or split)\n\
range of frequencies is desired.\n\
\n\
Currently only single stubs with an additional delay line are supported by\n\
this program. The structure looks as follows:\n\
\n\
         [LOAD/ANTENNA]\n\
               ||\n\
               ||\n\
               ||\n\
               ||  (delay line)\n\
               ||\n\
               ||\n\
               ||\n\
(T-connector)  ##===================[O/C]    ([O/C] = open or closed\n\
               ||       (stub)                        circuit at end)\n\
               ||\n\
               ||\n\
               ||\n\
               ||\n\
              ....  (line of unspecified length to generator/transceiver)\n\
               ||\n\
               ||\n\
               ||\n\
               ||\n\
     [GENERATOR/TRANSCEIVER]\n\
\n\
In order to calculate the length of the delay line and the length of the\n\
stub for an open or closed circuit, the complex impedance of the load has\n\
to be known for each frequency where an optimization of the standing wave\n\
ratio (SWR) is desired.\n\
\n\
The following parameters have to be entered when the program is started:\n\
\n\
- Cable impedance in     The impedance Z_0 of the delay and stub lines in\n\
  ohms:                  ohms (real number, usually 50 ohms for coaxial\n\
                         cables).\n\
\n\
- Cable velocity         The velocity factor of the delay and stub lines\n\
  factor:                (often 0.66 for coaxial cables).\n\
\n\
- Minimum and maximum    Cable attenuation for delay and stub lines. If\n\
  attenuation in dB      more than one frequency is considered later (i.e.\n\
  per 100m:              if sample count > 1), the minimum attenutation\n\
                         refers to the lowest considered frequency while\n\
                         the maximum attenuation refers to the highest\n\
                         considered frequency. Values in between are\n\
                         interpolated. If sample count is 1, minimum and\n\
                         maximum should be set to identical values.\n\
\n\
- Minimum and maximum    Minimum and maximum length of the delay line. This\n\
  shift length in        value can also be negative to indicate shortening\n\
  meters:                an existing line.\n\
\n\
- Minimum and maximum    Minimum and maximum length of the stub. Only\n\
  stub length in         positive numbers are allowed here.\n\
  meters:\n\
\n\
- Sample count:          Number of different frequencies where the load's\n\
                         complex impedance is known and where the SWR shall\n\
                         be optimized.\n\
\n\
And for each sample/frequency:\n\
\n\
- Frequency in MHz\n\
- Real part of impedance in ohms\n\
- Imaginary part of impedance in ohms\n\
\n\
Note that the program currently uses a brute-force algorithm with a fixed\n\
number of tries to find an initial optimal shift and stub length, which is\n\
then fine-tuned. It is thus advisable to always limit the minimum and\n\
maximum shift and stub lengths to a reasonable range.\n\
\n\
For each run of the program, two delay and stub lengths are calculated: one\n\
for an open circuit stub and one for a closed circuit stub. For each of\n\
those two variants, the resulting worst SWR for the sampled frequencies is\n\
shown.\n\
\n\
  vy 73 de DL9JBE\n";


/* include directives */

#include <complex.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>


/* physical constants */

#define LIGHTSPEED 299.792458


/* parametrization of brute-force search */

#define FIRST_RESOLUTION 1000
#define FINE_RESOLUTION 100
#define OVERLAP 3


/* I/O helper macros and functions */

#define interactive() isatty(fileno(stdin))

static void error(const char *msg) {
  fprintf(stderr, "Program abnormally terminated: %s\n", msg);
  exit(1);
}

static double read_real(const char *restrict fmt, ...) {
  va_list ap;
  double a;
  va_start(ap, fmt);
  if (interactive()) {
    vprintf(fmt, ap);
    printf(": ");
    fflush(stdout);
  }
  if (scanf("%lf", &a) != 1) error("Could not parse real number");
  va_end(ap);
  return a;
}

static double complex read_cplx(const char *restrict fmt, ...) {
  va_list ap;
  double a, b;
  va_start(ap, fmt);
  if (interactive()) {
    vprintf(fmt, ap);
    printf(":\nreal part> ");
    fflush(stdout);
  }
  if (scanf("%lf", &a) != 1) error("Could not parse real part of complex number");
  if (interactive()) {
    printf("imag part> ");
    fflush(stdout);
  }
  if (scanf("%lf", &b) != 1) error("Could not parse imaginary part of complex number");
  va_end(ap);
  return a + I*b;
}

static double complex read_int(const char *restrict fmt, ...) {
  va_list ap;
  int n;
  va_start(ap, fmt);
  if (interactive()) {
    vprintf(fmt, ap);
    printf(": ");
    fflush(stdout);
  }
  if (scanf("%i", &n) != 1) error("Could not parse integer number");
  va_end(ap);
  return n;
}


/* data structure for impedance sample */

struct sample {
  /* logarithm of frequency, needed for interpolating attenuation */
  double log_freq;
  /* reflection coefficient of load (antenna) */
  double complex refl;
  /* complex factor changing the reflection coefficient
     per meter towards generator (transceiver) */
  double complex effect;
};


/* conversion functions between impedance/admittance,
   reflection coefficient, and standing wave ratio (SWR) */

/* convert reflection coefficient to normalized impedance */
static double complex impedance(double complex g) {
  return (1+g) / (1-g);
}

/* convert reflection coefficient to normalized admittance */
static double complex admittance(double complex g) {
  return (1-g) / (1+g);
}

/* convert normalized impedance to reflection coefficient */
static double complex reflection_z(double complex z) {
  return (z-1) / (z+1);
}

/* convert normalized admittance to reflection coefficient */
#define reflection_y(y) admittance(y)

/* convert reflection coefficient to standing wave ratio (SWR) */
static double swr(double complex g) {
  return impedance(cabs(g));
}


/* functions regarding shift and stub lines */

/* calculation of change of reflection coefficient for a certain impedance
   sample (at a given frequency), shift and stub line lengths, and stub line
   termination (1=open or -1=closed) */
static double match(
  struct sample *s, double shiftlen, double stublen, int stubend
) {
  return swr(
    reflection_y(
      admittance(s->refl * cpow(s->effect, shiftlen)) +
      admittance(stubend * cpow(s->effect, stublen))
    )
  );
}

/* search optimal shift and stub length and print results to stdout */
static void search(
  struct sample *samples,
  int sample_count,
  double min_shiftlen, double max_shiftlen,
  double min_stublen,  double max_stublen,
  int stubend
) {
  double shiftlen_step, shiftlen, best_shiftlen;
  double stublen_step,  stublen,  best_stublen;
  double this_swr, best_swr, worst_swr;
  int i, j, k;
  double resolution = FIRST_RESOLUTION;
  int first = 1;
  if (stubend == 1)
    printf("Searching optimal parameters for open-circuited stub...");
  else if (stubend == -1)
    printf("Searching optimal parameters for short-circuited stub...");
  else abort();
  fflush(stdout);
  while (1) {
    shiftlen_step = (max_shiftlen - min_shiftlen) / (resolution-1);
    stublen_step  = (max_stublen  - min_stublen ) / (resolution-1);
    if (!first && !(shiftlen_step > 0 || stublen_step > 0)) break;
    best_swr = 1.0 / 0.0;
    for (i=0; i<resolution; i++) {
      shiftlen = min_shiftlen + i * shiftlen_step;
      for (j=0; j<resolution; j++) {
        stublen = min_stublen + j * stublen_step;
        worst_swr = 1.0;
        for (k=0; k<sample_count; k++) {
          this_swr = match(samples+k, shiftlen, stublen, stubend);
          if (!(this_swr <= worst_swr)) worst_swr = this_swr;
        }
        if (worst_swr < best_swr) {
          best_shiftlen = shiftlen;
          best_stublen  = stublen;
          best_swr = worst_swr;
        }
        if (!(stublen_step > 0)) break;
      }
      if (!(shiftlen_step > 0)) break;
    }
    if (first) {
      printf(" DONE.\nFine tuning");
    } else {
      printf(".");
    }
    fflush(stdout);
    if (best_shiftlen - OVERLAP*shiftlen_step > min_shiftlen)
      min_shiftlen = best_shiftlen - OVERLAP*shiftlen_step;
    if (best_shiftlen + OVERLAP*shiftlen_step < max_shiftlen)
      max_shiftlen = best_shiftlen + OVERLAP*shiftlen_step;
    if (best_stublen - OVERLAP*stublen_step > min_stublen)
      min_stublen = best_stublen - OVERLAP*stublen_step;
    if (best_stublen + OVERLAP*stublen_step < max_stublen)
      max_stublen = best_stublen + OVERLAP*stublen_step;
    resolution = FINE_RESOLUTION;
    first = 0;
  }
  printf(" DONE.\n\n");
  printf("Shift length: %.3f meters\n", min_shiftlen);
  if (stubend == 1)
    printf("Stub length: %.3f meters (open-circuited)\n", min_stublen);
  else if (stubend == -1)
    printf("Stub length: %.3f meters (short-circuited)\n", min_stublen);
  else abort();
  printf("Worst SWR: %.2f\n\n", best_swr);
  fflush(stdout);
}


/* main program */

int main(int argc, char **argv) {
  double complex z0;
  double vf;
  double min_att, max_att;
  double min_log_freq = 1.0/0.0;
  double max_log_freq = -1.0/0.0;
  double log_freq_span;
  int sample_count;
  struct sample *samples;
  double min_shiftlen, max_shiftlen;
  double min_stublen,  max_stublen;
  double this_swr, worst_swr = 1.0;
  int i;
  if (argc >= 2) {
    if (!strcmp(argv[1], "--help")) {
      printf("%s\n\n%s\n", version, help);
      exit(0);
    } else if (!strcmp(argv[1], "--license")) {
      printf("%s\n\n%s\n\n%s\n", version, copyright, license);
      exit(0);
    } else {
      fprintf(stderr, "%s\n\n", version);
      fprintf(stderr, "Use command line argument --help for help.\n");
      fprintf(stderr, "Use command line argument --license for license information.\n");
      fprintf(stderr, "For using the program, call it without any arguments.\n");
      fprintf(stderr, "\n");
      exit(1);
    }
  }
  /* print program name and version number */
  printf("%s\n\n", version);
  /* read data and calculate all samples[i].effect without cable loss */
  z0 = read_real("Cable impedance in ohms (e.g. 50 for RG-213U)");
  if (!(creal(z0) > 0)) error("Cable impedance is not positive");
  vf = read_real("Cable velocity factor (e.g. 0.66 for RG-213U)");
  if (!(vf > 0 && vf <= 1)) error("Velocity factor is not greater than 0 and smaller than or equal to 1");
  min_att = read_real("Minimum attenuation in dB per 100m (e.g. 0.5 or 0)");
  if (!(min_att >= 0)) error("Minimum attenuation is not positive or zero");
  max_att = read_real("Maximum attenuation in dB per 100m (e.g. 0.5 or 0)");
  if (!(max_att >= min_att)) error("Maximum attenuation is not greater than or equal to minimum attenuation");
  min_shiftlen = read_real("Minimum shift length in meters");
  max_shiftlen = read_real("Maximum shift length in meters");
  if (!(max_shiftlen >= min_shiftlen)) error("Maximum shift length not greater than or equal to minimum shift length");
  min_stublen = read_real("Minimum stub length in meters");
  if (!(min_stublen >= 0)) error("Minimum stub length is not positive or zero");
  max_stublen = read_real("Maximum stub length in meters");
  if (!(max_stublen >= min_stublen)) error("Maximum stub length is not greater than or equal to minimum stub length");
  sample_count = read_int("Sample count");
  if (sample_count < 1) error("Sample count is not positive");
  samples = calloc(sample_count, sizeof(*samples));
  if (!samples) error("Memory allocation error");
  for (i=0; i<sample_count; i++) {
    double freq = read_real("Frequency #%i in MHz", i+1);
    double log_freq;
    double complex z;
    if (!(freq > 0)) error("Frequency is not positive");
    log_freq = log(freq);
    if (log_freq < min_log_freq) min_log_freq = log_freq;
    if (log_freq > max_log_freq) max_log_freq = log_freq;
    samples[i].log_freq = log_freq;
    z = read_cplx("Impedance #%i in ohms", i+1);
    if (!(creal(z) > 0)) error("Real part of impedance is not positive");
    samples[i].refl = reflection_z(z / z0);
    samples[i].effect = cexp(-I * 4 * M_PI / LIGHTSPEED / vf * freq);
    this_swr = swr(samples[i].refl);
    if (!(this_swr <= worst_swr)) worst_swr = this_swr;
  }
  /* modify all samples[i].effect to include cable loss */
  log_freq_span = max_log_freq - min_log_freq;
  for (i=0; i<sample_count; i++) {
    double att;
    if (log_freq_span > 0) {
      double rel = (samples[i].log_freq - min_log_freq) / log_freq_span;
      att = (1.0-rel) * min_att + rel * max_att;
    } else {
      att = (min_att + max_att) / 2.0;
    }
    samples[i].effect *= pow(10, att / -1000);
  }
  /* print worst SWR without matching and call search functions for optimal
     shift and stub line lenghts (which also print their result to stdout) */
  printf("\nWorst SWR without matching: %.2f\n\n", worst_swr);
  search(
    samples, sample_count,
    min_shiftlen, max_shiftlen,
    min_stublen,  max_stublen,
    1
  );
  search(
    samples, sample_count,
    min_shiftlen, max_shiftlen,
    min_stublen,  max_stublen,
    -1
  );
  /* return with exit code 0 */
  return 0;
}
